/*
 Uses TinyGPSPlus library (Mikal Hart V1.0.3)
 GPSDO setup TP5 and communicate
 VK3DIP V1.0 25/08/2020
  Initial Version Set TP and Cut Through
 VK3DIP V1.1 20/10/2020
  Add set Nav Mode to Stationary to optimise TP stability.
 VK3DIP V2.0 25/03/2021 
  Add LCD and status etc.
 VK3DIP V2.1 30/03/2021 
  Add SI5351 and clean up code to reduce RAM usage
 VK3DIP V2.2 01/04/2021 
 rearange code some more to reduce RAM usage 
 VK3DIP V2.3 25/08/2022
 add extra two chars of locator (10 total), limit altitude to integer component.  
 V3 bypassed fork
 VK3DIP V4.1 26/08/2022
 Cleaned up code and adding setting of SI5351 via rotary encoder 
 VK3DIP V4.2 01/09/2022
 Made frequency commit optional, default auto Si5351 set.

 
  */

// *  INCLUDES
#include <EEPROM.h>
#include "si5351.h"
#include "Wire.h"
#include <SimpleRotary.h>
#include <LiquidCrystal_I2C.h>  
#include <SoftwareSerial.h> 
#include <TinyGPS++.h>


// Defining Const...

//======================================
// HARDWARE: Arduino I/O Pin allocations etc..

// Rotary Encoder...
const int  rotEncAPin = 6;
const int  rotEncBPin  =  7;
const int  rotEncSwPin = 5;

// Switches...
const int   modeSw1 = 8;

// Display...
const int   lcdCols = 16;
const int   lcdRows = 2;
const int   lcdAdd = 0x27;



// GPS Serial
const int gpsRX = 4;
const int gpsTX = 3;
const int gpsRate = 9600;


// EEPROM Addresses
// addresses 0 to 1023 for Nano
//1 for  each byte, 2 for each int, 4 for long etc.
// save Level (byte) and Freq in Hz (Unsigned Long) for each channel
// Channel 0
const int ch0LevA = 0; // Level (1 byte)
const int ch0FreqA = 1 ; // Freq 1,2,3,4 (4 bytes for a UL)
// Channel 1
const int ch1LevA = 5; // Level (byte)
const int ch1FreqA = 6 ; // Freq 6,7,8,9 (4 bytes for a UL)
// Channel 2
const int ch2LevA = 10; // Level (byte)
const int ch2FreqA = 11 ; // Freq 11,12,13,14 (4 bytes for a UL)


// Si5351 Defaults
// Reference Frequency set from GPSDO so 10MHz
const unsigned long refFreq = 10000000UL;
// Using nominal values stated in library doco, typical Si5351 can go higher or lower 
// but probably we need to manually set. 
// Maximum Frequency Setable
const unsigned long maxFreq = 150000000UL; // 150MHz
// Minimum Frequency Setable 
const unsigned long minFreq = 8000UL;  //8KHz
// Maximum Level Setable
const byte maxLev = 4;  



//Channel 0
const byte ch0LevD = 3; // 0 = off, 1 = 2mA, 2 = 4mA, 3 = 6mA, 4 =8mA
const unsigned long ch0FreqD = 22625000UL; // 22.625MHz

//Channel 1
const byte ch1LevD = 3; // 0 = off, 1 = 2mA, 2 = 4mA, 3 = 6mA, 4 =8mA
const unsigned long ch1FreqD = 15600000UL; // 15.6MHz

//Channel 2
const byte ch2LevD = 3; // 0 = off, 1 = 2mA, 2 = 4mA, 3 = 6mA, 4 =8mA
const unsigned long ch2FreqD = 30200000UL; // 30.2MHz



// end of Constants
//=================================================================


//Declare (and initialize) Global Variables...

// Holds current values for Levels and Freqs in Hz for each channel
// Channel 0
byte ch0LevV = ch0LevD; 
unsigned long ch0FreqV = ch0FreqD ;
// Channel 1
byte ch1LevV = ch1LevD; 
unsigned long ch1FreqV = ch1FreqD ;
// Channel 2
byte ch2LevV = ch2LevD; 
unsigned long ch2FreqV = ch2FreqD ;

byte selCur = 6; // Variable to hold cursor position of currently selected item 0-15 not all values do anything 
byte chanNum = 0; // Variable to hold currently selected channel 0, 1, or 2


// booleans to keep track of things
boolean i2c_found = false; // flag for if found SI5351
boolean gps_locked = false; // flag for if gps sees more than 3 sats ( assumed frequency locked)
boolean gps_serial = false; // flag for gps serial running (true)
boolean ch0_saved = false; //flag (true) to indicate current values for channel 0 saved to EEPROM
boolean ch1_saved = false; //flag (true) to indicate current values for channel 0 saved to EEPROM
boolean ch2_saved = false; //flag (true) to indicate current values for channel 0 saved to EEPROM
boolean suLn0_dirty = true; // flag (true) if first line of setup display needs to be written.
boolean suCh_dirty = true; // flag (true) if channel number on setup display needs to be written.
boolean suLv_dirty = true; // flag (true) if level on setup display needs to be written.
boolean suFreq_dirty = true; // flag (true) if frequency on setup display needs to be written.
boolean suSv_dirty = true; // flag (true) if saved state on setup display needs to be written.
boolean suCur_dirty = true; // flag (true) if menu cursor on setup display needs to be written.
boolean autoCommit = true; // if true any frequency change automatically sent to Si5351, if false wait for button push.

// end of Variables
//=================================================================

// * Instantiate Objects

// The LCD
LiquidCrystal_I2C lcd(lcdAdd,lcdCols,lcdRows);  

// The Software serial to the GPS module
SoftwareSerial GPS(gpsRX, gpsTX);

// The TinyGPS++ object
TinyGPSPlus gps;

// The Si5351
Si5351 si5351;

// The rotary encoder 
SimpleRotary rotary(rotEncAPin,rotEncBPin,rotEncSwPin);

// Setup everything *************************************************************************************    
void setup()
{

  // Start USB/Serial back to PC at same rate as GPS
  Serial.begin(gpsRate);
  Serial.println(" ");
  Serial.println(F("Neo-7+ GPSDO "));
  // Version Info

  Serial.println(F("VK3DIP V4.2 Sept. 2022")); 
 

   // Setup input pin
  pinMode(modeSw1, INPUT_PULLUP); // SW1 for GPSDO normal or Set Freq Mode
   // Set the trigger to be either a HIGH or LOW pin (Default: HIGH)
  // Note this sets all three pins to use the same state.
  rotary.setTrigger(HIGH);

  // Set the debounce delay in ms  (Default: 2)
  rotary.setDebounceDelay(5);

  // Set the error correction delay in ms  (Default: 200)
  rotary.setErrorDelay(250);

  // read the values from the EEPROM reseting to defaults if outside limits
  myReadEE();

 
 
  
  // Do LCD Setup
  lcd.init();
  lcd.clear();         
  lcd.backlight();      // Make sure backlight is on
  lcd.cursor(); // Turn Cursor on because we are going to use it.
    // Print startup message to the LCD.
  lcd.setCursor(0,0);   //Set cursor to character 0 on line 0
  lcd.print(F("VK3DIP GPSDO V4+")); 
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Start Setup....."));

  // Do GPS Setup  
  // Start GPS Serial 
   GPS.begin(gpsRate);  // will stop receive later via GPS.end();then re enable via another begin
   gps_serial = true;
   GPSSetup();

    
  // initialize the Si5351 and set frequencies
   lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
   lcd.print(F("Setup SI5351....")); //Using all 16chars so no clear needed
   Serial.println(F("Setup SI5351...."));
  
     
   i2c_found = si5351.init(SI5351_CRYSTAL_LOAD_0PF, refFreq, 0);
     if(i2c_found)
       {
          // Set all frequencies etc. to latest values
          mySetSI5351(0);
          mySetSI5351(1);  
          mySetSI5351(2);
       }
     else
       {
         Serial.println(F("Device not found on I2C bus!"));
       }
    
  // Setup done,
  Serial.println(F("Setup done ..."));
  Serial.println(" ");
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Set up complete ")); //Using all 16chars so no clear needed
  
} // end Setup ********************************************************************************************


 // Main loop just to service the GPS and or frequency setup**********************************************************************
void loop()
{
  lcd.clear(); 
  while(1)
  {
     byte i1 =  digitalRead(modeSw1);
     if ((!gps_locked) || (i1 != 0)) // if we are not locked or not requested setup do normal gps   
      {
     
       if(!gps_serial) 
        {
           Serial.println(F("restart serial"));
          GPS.begin(gpsRate); // if not running start listening
          gps_serial = true;
          lcd.clear(); 
          makeSUdirty(); //mark all of settup info as need to be rewritten if we go back
        }
    if(GPS.available())
    {
      // Read GPS Serial inputs (if available) and send them to encoder
      gps.encode(GPS.read());

      // check if we have enough to do anything.
      GotAny();
 
     } // end gps available  
      }
      else   // do frequency setup
      {    

        // stop GPS serial interupts etc. if we need to as we don't need it for settup
        if (gps_serial)
        {
          GPS.end(); // stop serial interupts
          gps_serial = false;
        } 
         switch (rotTest()) // check for input and action if necessary
        {
         case 1: //  CW      
            doIncValue();
            if ((autoCommit) && (selCur < 14)) mySetSI5351(chanNum) ; 
            // if autocommit is true and cursor not on save write to Si5351

            break;
         case 2: // CCW
           doDecValue();
           if ((autoCommit) && (selCur < 14)) mySetSI5351(chanNum) ; 
            // if autocommit is true and cursor not on save write to Si5351
           break;
         case 5: // just button press
           doAction();
           break;
         case 11: //  SCW      
            doIncCurs();
            break;
         case 12: //SCCW
           doDecCurs();
           break;
          default: // anything else
           // do nothing
            break; 
        }

        doSetupDisp(); // show the screen
       
      } // end Frequency settup
    } // end While
  } // end Loop  ********************************************************************************************


 
// ********************************Subs and Functions *******************************************************

// Read saved from EEPROM setting to defaults if outside ranges. *********************************************
void myReadEE(){
 Serial.println(F("Read Saved Values"));
   
 // Channel 0
  EEPROM.get(ch0LevA,ch0LevV); 
 if (ch0LevV > maxLev) ch0LevV = ch0LevD;
 EEPROM.get(ch0FreqA,ch0FreqV);
  if ((ch0FreqV > maxFreq) || (ch0FreqV < minFreq)) ch0FreqV = ch0FreqD;

 // Channel 1
  EEPROM.get(ch1LevA,ch1LevV);
 if (ch1LevV > maxLev) ch1LevV = ch1LevD;
 EEPROM.get(ch1FreqA,ch1FreqV);
 if ((ch1FreqV > maxFreq) || (ch1FreqV < minFreq)) ch1FreqV = ch1FreqD;

 // Channel 2
  EEPROM.get(ch2LevA,ch2LevV);
 if (ch2LevV > maxLev) ch2LevV = ch2LevD;
 EEPROM.get(ch2FreqA,ch2FreqV);
 if ((ch2FreqV > maxFreq) || (ch2FreqV < minFreq)) ch2FreqV = ch2FreqD;
     
} //end readEE ***************************************************************************************************

// rotTest Test Rotary Encoder and Button  for input*********************************************
byte rotTest(){
 byte i = rotary.rotate();
 if (i == 0) 
 { // No rotation
  if (rotary.push()== 1) i = 5; // Button pressed by itself
 }
 else 
 { // 1 CW or 2 CCW
  if ( digitalRead(rotEncSwPin) == 0)  i = i + 10; // Button is down indicate shift 
 }

 return i;  // returns 0= no input, 1 = CW, 2 = CCW, 5 = button pressed by itself, 11 = ShiftCW (pressed while rotated), 12 = ShiftCCW     
 
} //end rotTest ***************************************************************************************************

// makeSUdirty , make all settup info need to be rewritten *********************************
void makeSUdirty() {

suLn0_dirty = true; // flag (true) if first line of setup display needs to be written.
suCh_dirty = true; // flag (true) if channel number on setup display needs to be written.
suLv_dirty = true; // flag (true) if level on setup display needs to be written.
suFreq_dirty = true; // flag (true) if frequency on setup display needs to be written.
suSv_dirty = true; // flag (true) if saved state on setup display needs to be written.
suCur_dirty = true; // flag (true) if menu cursor on setup display needs to be written.

} // end makeSUdirty **********************************************************************

// Print number (unsigned long in this case) to say LCD with a fixed width field (default space paded) *************************
// width should always be greater than max needed and number will be right justified in it.************************
void printFixedWidthUL(Print &out, unsigned long number, byte width, char filler = ' ')
{
  unsigned long temp = number;
   
  //see how wide the number is
  while(temp && width){
    temp /= 10;
    width--;
  }
 
  //start by printing the rest of the width with pad char
  while(width){
    out.print(filler);
    width--;
  }

 // finally print the number
  out.print(number);
} 
//end printFixedWidthUL ***************************************************************************************************

// Do 16*2 LCD display as for Si5351 channel setup with cursor showing above currently selected item***********************
void doSetupDisp(){

 if (suLn0_dirty)  // do first line if necessary
 {
  lcd.clear(); //clear the screen if line 0 needs to be refreshed
  lcd.setCursor(0,0);   //Move cursor to character 0 on line 0
  lcd.print(F("ChLv Freq(Hz) Sv")); 
  suLn0_dirty = false;
 }

  if (suCh_dirty) // do channel Number if necessary
  {
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(chanNum);
  suCh_dirty = false;
  }

  if (suLv_dirty) // do level if necessary
  {
    lcd.setCursor(2,1);   //Move cursor to character 2 on line 1
   switch (chanNum)
    {
    case 0:
    lcd.print(ch0LevV);
    break;
    case 1:
    lcd.print(ch1LevV);
    break;
    case 2:
    lcd.print(ch2LevV);
    break;
    default:
    // do nothing
    break;
    }
   suLv_dirty = false;
  }
  
 if (suFreq_dirty) // do frequency if necessary
  {
    lcd.setCursor(4,1);   //Move cursor to character 4 on line 1
   switch (chanNum)
    {
    case 0:
    printFixedWidthUL(lcd,ch0FreqV,9);
    break;
    case 1:
    printFixedWidthUL(lcd,ch1FreqV,9);
    break;
    case 2:
    printFixedWidthUL(lcd,ch2FreqV,9);
    break;
    default:
    // do nothing
    break;
    }
   suFreq_dirty = false;
  }

 if (suSv_dirty) // do Saved State if necessary
  {
   checkSaved(); // check what is saved and set flags
   lcd.setCursor(14,1);   //Move cursor to character 14 on line 1
   switch (chanNum)
    {
    case 0:
    lcd.print(ch0_saved);
    break;
    case 1:
    lcd.print(ch1_saved);
    break;
    case 2:
    lcd.print(ch2_saved);
    break;
    default:
    // do nothing
    break;
    }
   suSv_dirty = false;
  }

   lcd.setCursor(selCur,1); // always set the cursor to current position on lower line

     
} //end doSetupDisp() ***********************************************************

// checkSaved  checks to see if current values equal saved values, sets chn_saved flags**************
void checkSaved(){

byte tempb;
unsigned long tempUL;
ch0_saved = false;
ch1_saved = false;
ch2_saved = false;

EEPROM.get(ch0LevA,tempb);
EEPROM.get(ch0FreqA,tempUL);
if ((tempb == ch0LevV)&&(tempUL == ch0FreqV)) ch0_saved = true;

EEPROM.get(ch1LevA,tempb);
EEPROM.get(ch1FreqA,tempUL);
if ((tempb == ch1LevV) && (tempUL == ch1FreqV))  ch1_saved = true;

EEPROM.get(ch2LevA,tempb);
EEPROM.get(ch2FreqA,tempUL);
if ((tempb == ch2LevV) && (tempUL == ch2FreqV)) ch2_saved = true;

     
} //end checkSaved ***********************************************************

// doEEsave  saves(updates if necessary) the values for current channel via put **************
void doEEsave(byte mych){

 switch (mych)
  {
    case 0:
    EEPROM.put(ch0LevA,ch0LevV);
    EEPROM.put(ch0FreqA,ch0FreqV);
      break;
   case 1:
    EEPROM.put(ch1LevA,ch1LevV);
    EEPROM.put(ch1FreqA,ch1FreqV);
      break;
   case 2:
    EEPROM.put(ch2LevA,ch2LevV);
    EEPROM.put(ch2FreqA,ch2FreqV);
      break; 
   default:
  //Do Nothing
      break;  
  }   
} //end doEEsave ***********************************************************

// doIncCurs() // increment the cursor **************
void doIncCurs(){

selCur = selCur + 1;
if (selCur > 15) selCur = 0;
suCur_dirty = true;

} //end doIncCurs ***********************************************************

// doDecCurs() // decrement the cursor **************
void doDecCurs(){
if (selCur == 0)
{
  selCur = 15;
}
else
{
  selCur = selCur - 1;
}
suCur_dirty = true;
} //end doDecCurs ***********************************************************

// doAction() // not turning but knob button pressed **************
void doAction(){
  
mySetSI5351(chanNum); // always set current values (level and frequency) to current channel of Si5351
if (selCur == 14) 
{
  doEEsave(chanNum);       // also if in position 14 do an EEPROM save of current channel values
  suSv_dirty = true;
}
 
} //end doAction ***********************************************************

// doIncValue increments the item under the cursor  **************
void doIncValue(){

 switch (selCur)
  {
   case 0: // its the channel number
      chanNum = chanNum + 1;
      if(chanNum > 2) chanNum = 0;
      suCh_dirty = true;  // most things will need to change if channel changes
      suLv_dirty = true; 
      suFreq_dirty = true; 
      suSv_dirty = true; 
      break;
   case 2: // its the level
      switch (chanNum)
      {
       case 0:
         ch0LevV = ch0LevV + 1;
         if(ch0LevV > 4) ch0LevV = 0;
         break;
       case 1:
         ch1LevV = ch1LevV + 1;
         if(ch1LevV > 4) ch1LevV = 0;
         break;
       case 2:
         ch2LevV = ch2LevV + 1;
         if(ch2LevV > 4) ch2LevV = 0;
         break;
       default:
        //Do Nothing
        break;  
      }  
      suLv_dirty = true; // level and possibly saved state need rewriting
      suSv_dirty = true;  
      break;
   case 4: // its the frequency
      incFreq(chanNum,100000000UL);
      break; 
   case 5:
      incFreq(chanNum,10000000UL);
      break;
   case 6:
      incFreq(chanNum,1000000UL);
      break;
   case 7:
      incFreq(chanNum,100000UL);
      break; 
   case 8:
      incFreq(chanNum,10000UL);
      break;
   case 9:
      incFreq(chanNum,1000UL);
      break; 
   case 10:
      incFreq(chanNum,100UL);
      break;
   case 11:
      incFreq(chanNum,10UL);
      break;
   case 12:
      incFreq(chanNum,1UL);
      break; 
   default:
  // all other cases do nothing
      break;  
  }   
} //end doIncValue ***********************************************************

// incFreq adds step to freq for selected channel  **************
void incFreq(byte mych, unsigned long fStep){

    switch (mych)
    {
     case 0:
        ch0FreqV = ch0FreqV + fStep;
    if(ch0FreqV > maxFreq) ch0FreqV = maxFreq;
       break;
     case 1:
        ch1FreqV = ch1FreqV + fStep;
    if(ch1FreqV > maxFreq) ch1FreqV = maxFreq;
       break;
     case 2:
        ch2FreqV = ch2FreqV + fStep;
    if(ch2FreqV > maxFreq) ch2FreqV = maxFreq;
       break;
     default:
    //Do Nothing
      break;  
    }   
   suFreq_dirty = true; 
   suSv_dirty = true;
} //end incFreq ***********************************************************

// doDecValue increments the item under the cursor  **************
void doDecValue(){

 switch (selCur)
  {
   case 0: // its the channel number
      if(chanNum== 0)
       {
         chanNum = 2;
       }
      else
       {
        chanNum = chanNum - 1;  
       }
      suCh_dirty = true;  // most things will need to change if channel changes
      suLv_dirty = true; 
      suFreq_dirty = true; 
      suSv_dirty = true; 
      break;
   case 2: // its the level
  switch (chanNum)
    {
     case 0:
      if(ch0LevV == 0)
       {
         ch0LevV = 4;
       }
      else
       {
        ch0LevV = ch0LevV - 1;  
       }
       break;
     case 1:
      if(ch1LevV== 0)
       {
         ch1LevV = 4;
       }
      else
       {
        ch1LevV = ch1LevV - 1;  
       }
       break;
      case 2:
      if(ch2LevV== 0)
       {
         ch2LevV = 4;
       }
      else
       {
        ch2LevV = ch2LevV - 1;  
       }
       break;
 
     default:
    //Do Nothing
      break;  
    } 
    suLv_dirty = true; // level and possibly saved state need rewriting
    suSv_dirty = true;    
    break;
   case 4: // Frequency ones
      decFreq(chanNum,100000000UL);
      break; 
   case 5:
      decFreq(chanNum,10000000UL);
      break;
   case 6:
      decFreq(chanNum,1000000UL);
      break;
   case 7:
      decFreq(chanNum,100000UL);
      break; 
   case 8:
      decFreq(chanNum,10000UL);
      break;
   case 9:
      decFreq(chanNum,1000UL);
      break; 
   case 10:
      decFreq(chanNum,100UL);
      break;
   case 11:
      decFreq(chanNum,10UL);
      break;
   case 12:
      decFreq(chanNum,1UL);
      break; 
   default:
     // all other cases do nothing
      break;  
  }   
} //end doDecValue ***********************************************************

// decFreq subtracts step from freq for selected channel  **************
void decFreq(byte mych, unsigned long fStep){

    switch (mych)
    {
     case 0:
        if(fStep > ch0FreqV) // would take it negative
        {
         ch0FreqV = minFreq;
         }
        else
        {
         ch0FreqV = ch0FreqV - fStep;
         if(ch0FreqV < minFreq) ch0FreqV = minFreq;
        }    
        break;
     case 1:
        if(fStep > ch1FreqV) // would take it negative
        {
         ch1FreqV = minFreq;
         }
        else
        {
         ch1FreqV = ch1FreqV - fStep;
         if(ch1FreqV < minFreq) ch1FreqV = minFreq;
        }    
        break;
     case 2:
       if(fStep > ch2FreqV) // would take it negative
        {
         ch2FreqV = minFreq;
         }
        else
        {
         ch2FreqV = ch2FreqV - fStep;
         if(ch2FreqV < minFreq) ch2FreqV = minFreq;
        }    
        break;
     default:
    //Do Nothing
      break;  
    }  
   suFreq_dirty = true; 
   suSv_dirty = true; 
   
} //end decFreq ***********************************************************




// GPS Setup ************************************************************************************************
void GPSSetup(){
  // 
  // 
  uint8_t UBXNav[] = {
    0xB5, 0x62, 0x06, 0x24, 0x24, 0x00, 0xFF, 0xFF, 0x02, 0x03, 0x00, 0x00, 0x00, 0x00, 0x10, 0x27,
    0x00, 0x00, 0x05, 0x00, 0xFA, 0x00, 0xFA, 0x00, 0x64, 0x00, 0x2C, 0x01, 0x00, 0x3C, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x4E, 0x60 }; 
  
   //  This UBX command sets the Navigation Mode to Stationary which is optimal for accuracy of Time Pulse
  Serial.println(F("Setting uBlox Nav mode: "));
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Set up Nav mode.")); //Using all 16chars so no clear needed
  RepSendUBXBuf(UBXNav,sizeof(UBXNav)/sizeof(uint8_t));

 //  This UBX command sets the TimePulse to 1HZ unlocked, 100KHz Locked, both 50% duty cycle
  Serial.println(F("Initialize TP5 to 100KHz ..."));  
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Init TP = 100KHz")); //Using all 16chars so no clear needed
   uint8_t UBXTP[] = {
    0xB5, 0x62, 0x06, 0x31, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00, 0x32, 0x00, 0x00, 0x00, 0x01, 0x00,
    0x00, 0x00, 0xA0, 0x86, 0x01, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00,
    0x00, 0x00, 0xEF, 0x00, 0x00, 0x00, 0xA1, 0xBA };
   RepSendUBXBuf(UBXTP,sizeof(UBXTP)/sizeof(uint8_t));

 // get rid of all normal messages except GGA
  Serial.println(F("Turn off all messages save GGA "));
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Enable GGA only ")); //Using all 16chars so no clear needed
  
   Serial.println(F("Switching off NMEA GLL: "));
  uint8_t UBXGLL[] = {
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x2B };
   RepSendUBXBuf(UBXGLL,sizeof(UBXGLL)/sizeof(uint8_t));

   Serial.println(F("Switching off NMEA GSA: "));
    uint8_t UBXGSA[] = {
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x32 };
  RepSendUBXBuf(UBXGSA,sizeof(UBXGSA)/sizeof(uint8_t));

   Serial.println(F("Switching off NMEA GSV: "));
    uint8_t UBXGSV[] = {
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x39 };
   RepSendUBXBuf(UBXGSV,sizeof(UBXGSV)/sizeof(uint8_t));

   Serial.print(F("Switching off NMEA RMC: "));
   uint8_t UBXRMC[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, 0x40 };
   RepSendUBXBuf(UBXRMC,sizeof(UBXRMC)/sizeof(uint8_t));

   Serial.print(F("Switching off NMEA VTG: "));
   uint8_t UBXVTG[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x05, 0x47                   };
   RepSendUBXBuf(UBXVTG,sizeof(UBXVTG)/sizeof(uint8_t));
  
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Now GGA only    ")); //Using all 16chars so no clear needed  
} // end GPSSetup ******************************************************************************************************


// keep sending UBX message until we get +ve ack ***********************************************************************
// allows for possible serial problems
void RepSendUBXBuf(uint8_t *MSG, uint8_t len){
   byte gps_set_sucess=0;
   while(!gps_set_sucess)
  {
    for(int i=0; i<len; i++) {
      GPS.write(MSG[i]);
      if ( i % 10 == 0 && i != 0) Serial.println() ;
      PrnHEX(MSG[i]);
    }
  GPS.println();
  gps_set_sucess=getUBX_ACK(MSG);
  }
 } //end RepSendUBXBuf  ************************************************************************************************

// Calculate expected UBX ACK packet and parse UBX response from GPS ***************************************************
boolean getUBX_ACK(uint8_t *MSG) {
  uint8_t b;
  uint8_t ackByteID = 0;
  uint8_t ackPacket[10];
  unsigned long startTime = millis();
  Serial.println();
  Serial.print("Reading ACK response: ");
 Serial.println();
  // Construct the expected ACK packet    
  ackPacket[0] = 0xB5;  // header
  ackPacket[1] = 0x62;  // header
  ackPacket[2] = 0x05;  // class
  ackPacket[3] = 0x01;  // id
  ackPacket[4] = 0x02;  // length
  ackPacket[5] = 0x00;
  ackPacket[6] = MSG[2];  // ACK class
  ackPacket[7] = MSG[3];  // ACK id
  ackPacket[8] = 0;   // CK_A
  ackPacket[9] = 0;   // CK_B
 
  // Calculate the checksums
  for (uint8_t i=2; i<8; i++) {
    ackPacket[8] = ackPacket[8] + ackPacket[i];
    ackPacket[9] = ackPacket[9] + ackPacket[8];
  }
 
  while (1) {
 
    // Test for success
    if (ackByteID > 9) {
      // All packets in order!
      Serial.println(" (SUCCESS!)");
      return true;
    }
 
    // Timeout if no valid response in 3 seconds
    if (millis() - startTime > 3000) { 
      Serial.println(" (FAILED!) Try Again");
      return false;
    }
 
    // Make sure data is available to read
    if (GPS.available()) {
      b = GPS.read();
 
      // Check that bytes arrive in sequence as per expected ACK packet
      if (b == ackPacket[ackByteID]) { 
        ackByteID++;
        PrnHEX(b);
      } 
      else {
        ackByteID = 0;  // Reset and look again, invalid order
      }
 
    }
  }
} // end getUBX_ACK ***************************************************************************************************

// Print Hex Chars nicely. ********************************************************************************************
 void PrnHEX(uint8_t MyChr) {
 
    Serial.print("0x");
    Serial.print(MyChr < 16 ? "0" : "");
    Serial.print(MyChr, HEX);
    Serial.print(" ");
    
} //end PrnHEX ********************************************************************************************************


// Check for any decoded data and action if present *******************************************************************
void GotAny() {
     
   if (gps.location.isUpdated()) // if enough for a location
       {   
         
         lcd.setCursor(0,1);   //Set cursor to character 0 on line 1
         char myloc[11]; //array to hold locator string
         calcLocator(myloc, gps.location.lat(), gps.location.lng());
         lcd.print(myloc);
       }

    if (gps.altitude.isUpdated()) // got altitude
       {
         lcd.setCursor(11,1);   //Set cursor to character 9 on line 1
         lcd.print(int(gps.altitude.meters()));
         lcd.print(F("m ")); 
       }

    if (gps.satellites.isUpdated()) // got satellites
       {
        
        lcd.setCursor(9,0);   //Set cursor to character 9 on line 0
        lcd.print(F("  "));  // clean out to allow for less than 10 satelites
        lcd.setCursor(9,0);   //Set cursor to character 9 on line 0
        int i2 = gps.satellites.value();
        if (i2 > 2)
        {
          gps_locked = true;
        }
        else
        {
          gps_locked = false;
        }
        lcd.print(i2);
        
       }

    if (gps.hdop.isUpdated()) // got HDOP
       {
        lcd.setCursor(12,0);   //Set cursor to character 12 on line 0
       lcd.print(gps.hdop.hdop());
      
       }

      if (gps.time.isUpdated()) // got Time
       {
        lcd.setCursor(0,0);   //Set cursor to character 0 on line 0
        MyUTC(); // LCD print time
        // park the cursor out the way
        lcd.setCursor(15,1);   //Set cursor to character 15 on line 1
       } 
       

      
}// end GotAny ********************************************************************************************************

// Send Good Time to LCD at current cursor ****************************************************************************

void MyUTC() {
   if (gps.time.isValid())
  {
    if (gps.time.hour() < 10) lcd.print(F("0"));
    lcd.print(gps.time.hour());
    lcd.print(F(":"));
    if (gps.time.minute() < 10) lcd.print(F("0"));
    lcd.print(gps.time.minute());
    lcd.print(F(":"));
    if (gps.time.second() < 10) lcd.print(F("0"));
    lcd.print(gps.time.second());
    //lcd.print(F("."));
    //if (gps.time.centisecond() < 10) lcd.print(F("0"));
    //lcd.print(gps.time.centisecond());
  }
  else
  {
    lcd.print(F("INVALID "));
  }
}  // end MyUTC  ******************************************************************************************************

// calculate maidenhead gridsquare from given latitude and longitude **************************************************
// returns 8 char grid locator in elsewhere declared 9 char array loc passed by *pointer

void calcLocator(char *loc, double latitude, double longitude) {
  // Latitude and longitude are in decimal degrees
  // *loc is a reference to a char array declared global elsewhere
  int tempint; // int variable to hold individual chars
  double remainder; // keep track of remainder
 
  // do longitude calcs
  remainder = longitude + 180.0;
  tempint = (int)(remainder / 20.0);
  remainder = remainder - (double)tempint * 20.0;
  loc[0] = (char)tempint + 'A';
   
  tempint = (int)(remainder / 2.0);
  remainder = remainder - 2.0 * (double)tempint;
  loc[2] = (char)tempint + '0';
   
  tempint = (int)(12.0 * remainder);
  remainder = remainder - (double)tempint / 12;
  loc[4] = (char)tempint + 'a';
  
  tempint = (int)(remainder *120);
  remainder = remainder - (double)tempint / 120;
  loc[6] = (char)tempint + '0';

  tempint = (int)(remainder *2880);
  loc[8] = (char)tempint + 'a';
  
  // do latitude calcs
  remainder = latitude + 90.0;
  tempint = (int)(remainder / 10.0);
  remainder = remainder - (double)tempint * 10.0;
  loc[1] = (char)tempint + 'A';
  
  tempint = (int)(remainder);
  remainder = remainder - (double)tempint;
  loc[3] = (char)tempint + '0';
  
  tempint = (int)(24.0 * remainder);
  remainder = remainder - (double)tempint / 24;
  loc[5] = (char)tempint + 'a';
  
  tempint = (int)(remainder * 240);
  remainder = remainder - (double)tempint / 240;
  loc[7] = (char)tempint + '0';

  tempint = (int)(5760.0 * remainder);
 
  loc[9] = (char)tempint + 'a';
  
  //terminate result
   loc[10] = (char)0;
} // end calcLocator *********************************************************************************************************

// Turn Uint32_t into Uint64_t ( unsigned long in Hertz to unsigned long long in 0.01Hz suitable for SI5351 set***************
 unsigned long long ToCHz(uint32_t MyHz) {
 
   return MyHz * SI5351_FREQ_MULT ;
    
} //end ToHz ********************************************************************************************************

// Set frequencies for the SI5351 ********************************************************************************************
// Assume called only after 10MHz ref is available.
void mySetSI5351(byte mych) {
      if(i2c_found)
       {
        switch (mych) 
        {
          case 0:
            si5351.set_freq(ToCHz(ch0FreqV), SI5351_CLK0);
            switch (ch0LevV)
            {
              case 0:
                si5351.output_enable(SI5351_CLK0, 0); //disable output
                break;
              case 1:
                si5351.output_enable(SI5351_CLK0, 1); //enable output
                si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_2MA);
                break;
              case 2:
                si5351.output_enable(SI5351_CLK0, 1); //enable output
                si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_4MA);
                break;
              case 3:
                si5351.output_enable(SI5351_CLK0, 1); //enable output
                si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_6MA);
                break;
              case 4:
                si5351.output_enable(SI5351_CLK0, 1); //enable output
                si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_8MA);
                break;
              default:
              // Do nothing
                break;
            }
            break;
          case 1:
            si5351.set_freq(ToCHz(ch1FreqV), SI5351_CLK1);
             switch (ch1LevV)
            {
              case 0:
                si5351.output_enable(SI5351_CLK1, 0); //disable output
                break;
              case 1:
                si5351.output_enable(SI5351_CLK1, 1); //enable output
                si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_2MA);
                break;
              case 2:
                si5351.output_enable(SI5351_CLK1, 1); //enable output
                si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_4MA);
                break;
              case 3:
                si5351.output_enable(SI5351_CLK1, 1); //enable output
                si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_6MA);
                break;
              case 4:
                si5351.output_enable(SI5351_CLK1, 1); //enable output
                si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_8MA);
                break;
              default:
              // Do nothing
                break;
            }
            break;
          case 2:
            si5351.set_freq(ToCHz(ch2FreqV), SI5351_CLK2);
             switch (ch2LevV)
            {
              case 0:
                si5351.output_enable(SI5351_CLK2, 0); //disable output
                break;
              case 1:
                si5351.output_enable(SI5351_CLK2, 1); //enable output
                si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_2MA);
                break;
              case 2:
                si5351.output_enable(SI5351_CLK2, 1); //enable output
                si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_4MA);
                break;
              case 3:
                si5351.output_enable(SI5351_CLK2, 1); //enable output
                si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_6MA);
                break;
              case 4:
                si5351.output_enable(SI5351_CLK2, 1); //enable output
                si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_8MA);
                break;
              default:
              // Do nothing
                break;
            }
            break;
          default:
         // Do nothing because it shouldn't be here
            break;
        }  
         delay(500); // delay a bit to stabalize          
       }

 
} //end MySetSI5351  **********************************************************************************************************
 
